unit FIN;

interface

uses UcCompressorMisc,windows;

type
    TTempBuffer = array[1..32768] of byte;
    PTempBuffer = ^TTempBuffer;

    //TMyBuffer = array[0..32768-1] of byte; //32k


var
   InBuffer, OutBuffer: PTempBuffer;
   Bytesleft: Cardinal; //the bytes left to be read for compress or decompress
   FileCompressedSize: Cardinal;



//stream
procedure FIN_InitCompression(var Source: UcStream; var Destination: UcStream);
procedure FIN_FreeCompression(var Source: UcStream; var Destination: UcStream);
procedure FIN_CompressStream(var Source: UcStream; var Destination: UcStream);
procedure FIN_InitDecompression(var Source: UcStream; var Destination: UcStream);
procedure FIN_FreeDecompression(var Source: UcStream; var Destination: UcStream);
procedure FIN_DecompressStream(var Source: UcStream; var Destination: UcStream);

implementation

{****************************** Encode/Decode FIN ******************************}
type
    tpcTable = array[0..32768] of byte;
    ppcTable = ^TpcTable;

var
   //compress
   p1, p2: byte;             // previous 2 characters
   buf: array[0..8] of byte; // keeps characters temporarily
   ctr: integer;             // number of characters in mask
   bctr: integer;            // position in buf
   mask: byte;               // mask to mark successful predictions
   pcTable: ppcTable;

   //decompress
   ci,co: byte;     // characters (in and out)
   //p1,p2: shortint; // previous 2 characters
   //ctr: integer;    // number of characters processed for this mask
   //mask: byte;      // mask to mark successful predictions


function INDEX(var p1,p2): integer;
begin
   Result := ( Byte(p1) shl 7) xor Byte(p2);
end;

procedure FIN_InitCompression(var Source: UcStream; var Destination: UcStream);
begin
     p1 := 0;
     p2 := 0;
     ctr := 0;
     bctr := 0;
     mask := 0;
     New(pcTable);
     New(InBuffer);
     New(OutBuffer);
     FillChar(pcTable^,32768,' ');
     GetMem(Source.Buffer,32768);
     GetMem(Destination.Buffer,32768); //the compressed data must be small than the original data
     Source.MaxSize := 32768;
     Destination.MaxSize := 32768;
end;

procedure FIN_FreeCompression(var Source: UcStream; var Destination: UcStream);
begin
     Dispose(pcTable);
     Dispose(InBuffer);
     Dispose(OutBuffer);
     FreeMem(Source.Buffer);
     FreeMem(Destination.Buffer);
end;

procedure FIN_CompressStream(var Source: UcStream; var Destination: UcStream);
var
   c: byte;                  // character
   i: integer;               // loop counter
  { p1, p2: byte;             // previous 2 characters
   buf: array[0..8] of byte; // keeps characters temporarily
   ctr: integer;             // number of characters in mask
   bctr: integer;            // position in buf
   mask: byte;               // mask to mark successful predictions
   }
   bytesread: integer;
   srcpos,destpos: cardinal;
   //pcTable: ppcTable;
   procedure WriteToBuffer(chr: Byte);
   begin

        inc(Destination.Size);
        if Destination.Size >= Source.Size then //this block is no possible to compress than don't save it to the buffer
           begin
           end
        else
            begin
                 Destination.Buffer^[destpos] := chr;
                 inc(destpos);
            end;
   end;
   procedure ReadFromBuffer(var chr: byte; var numbytes: integer);
   begin
        if srcpos > Source.Size-1 then
           numbytes := 0
        else
            begin
                 chr := Source.Buffer^[srcpos];
                 inc(srcpos);
                 bytesread := 1;
            end;
   end;
begin
   BytesLeft := Source.Size;// BytesToCompress; //we need to compress 'BytesToCompress' bytes
     Destination.Size := 0; //the file compressed size for now is 0
     srcpos := 0;
     destpos := 0;

     ReadFromBuffer(c,bytesread);
     while bytesread <> 0 do
     begin
          // try to predict the next character
          if pcTable^[INDEX(p1,p2)] = c then
             // correct prediction, mark bit for correct prediction
             mask := mask xor (1 shl ctr)
          else
              begin
                   // wrong prediction, but next time ...
                   pcTable^[INDEX(p1,p2)] := c;
                   // buf keeps character temporarily in buffer
                   buf[bctr] := c;
                   inc(bctr);
              end;
          // test if mask is full (8 characters read)
          inc(ctr);
          if ctr= 8 then
             begin
                  // write mask
                  WriteToBuffer(mask);
                  // write kept characters
                  for i:=0 to bctr-1 do
                      WriteToBuffer(buf[i]);
                  // reset variables
                  ctr :=0;
                  bctr:=0;
                  mask:=0;
             end;
          // shift characters
          p1 := p2;
          p2 := c;
          ReadFromBuffer(c,bytesread);
     end;
     if Source.IsFinalPiece then //if its the last piece the  clear the buffer
        begin
             // EOF, but there might be some left for output
             if ctr >= 1 then
                begin
                     // write mask
                     WriteToBuffer(mask);
                     // write kept characters
                     for i:=0 to bctr-1 do
                         WriteToBuffer(buf[i]);
                end;
        end;
end;

procedure FIN_InitDecompression(var Source: UcStream; var Destination: UcStream);
begin
     p1 := 0;
     p2 := 0;
     ctr := 8;
     mask := 0;
     New(InBuffer);
     New(OutBuffer);
     New(pcTable);
     fillchar(pcTable^, 32768, ' '); // space (ASCII 32) is the most used char
     GetMem(Source.Buffer,32768);
     GetMem(Destination.Buffer,32768);
     Source.MaxSize := 32768;
     Destination.MaxSize := 32768;
end;

procedure FIN_FreeDecompression(var Source: UcStream; var Destination: UcStream);
begin
     Dispose(pcTable);
     Dispose(InBuffer);
     Dispose(OutBuffer);
     FreeMem(Source.Buffer);
     FreeMem(Destination.Buffer);
end;


procedure FIN_DecompressStream(var Source: UcStream; var Destination: UcStream);
var
{   ci,co: byte;     // characters (in and out)
   p1,p2: shortint; // previous 2 characters
   ctr: integer;    // number of characters processed for this mask
   mask: byte;      // mask to mark successful predictions
}
   bytesread: integer;
   //pcTable: ppcTable;
   decompressedsize:integer;

   destpos,srcpos:integer;
   procedure WriteToBuffer(chr: Byte);
   begin
        Destination.Buffer^[destpos] := chr;
        inc(destpos);
        inc(decompressedsize);
   end;
   procedure ReadFromBuffer(var chr: byte; var numbytes: integer);
   begin
        if srcpos > Source.Size-1 then
           numbytes := 0
        else
            begin
                 chr := Source.Buffer^[srcpos];
                 inc(srcpos);
                 numbytes := 1;
            end;
   end;
begin
     srcpos := 0;
     destpos := 0;
     decompressedsize := 0;
     //BytesLeft := BytesToExtract;//we need to compress 'BytesToExtract' bytes
     ReadFromBuffer(ci,bytesread);
     while bytesread <> 0 do
     begin
          // get mask (for 8 characters)
          mask := byte(shortint(ci));
          // for each bit in the mask
          for ctr:=0 to 7 do
              begin
                   if (mask and (1 shl ctr)) >= 1 then
                      // predicted character
                      co := pcTable^[INDEX(p1,p2)]
                   else
                       begin
                            // not predicted character
                            ReadFromBuffer(co,bytesread);
                            if bytesread = 0 then
                            begin
                                 //WriteBuffer(true,1);// decompression completed! Clean the buffer and exit
                                 destination.Size := decompressedsize;
                                 Exit;
                            end;
	                    pcTable^[INDEX(p1,p2)] := byte(co);
                       end;
                   WriteToBuffer(co);
                   p1 := p2;
                   p2 := co;
              end;
          ReadFromBuffer(ci,bytesread);
   end;
   destination.Size := decompressedsize;
end;

end.

